﻿using System;
using System.ComponentModel;
using System.IO;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Markup;
using System.Windows.Media.Animation;
using System.Windows.Media.Imaging;

using TestLibrary.Decoding;
using TestLibrary.Extension;

namespace TestLibrary
{
    /// <summary>
    /// 애니메이션 동작
    /// </summary>
    public static class AnimationBehavior
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Routed Event
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 에러 이벤트 - ErrorEvent

        /// <summary>
        /// 에러 이벤트
        /// </summary>
        public static readonly RoutedEvent ErrorEvent = EventManager.RegisterRoutedEvent
        (
            "Error",
            RoutingStrategy.Bubble,
            typeof(AnimationErrorEventHandler),
            typeof(AnimationBehavior)
        );

        #endregion
        #region 다운로드 진행 이벤트 - DownloadProgressEvent

        /// <summary>
        /// 다운로드 진행 이벤트
        /// </summary>
        public static readonly RoutedEvent DownloadProgressEvent = EventManager.RegisterRoutedEvent
        (
            "DownloadProgress",
            RoutingStrategy.Bubble,
            typeof(DownloadProgressEventHandler),
            typeof(AnimationBehavior)
        );

        #endregion
        #region 로드시 이벤트 - LoadedEvent

        /// <summary>
        /// 로드시 이벤트
        /// </summary>
        public static readonly RoutedEvent LoadedEvent = EventManager.RegisterRoutedEvent
        (
            "Loaded",
            RoutingStrategy.Bubble,
            typeof(RoutedEventHandler),
            typeof(AnimationBehavior)
        );

        #endregion
        #region 애니메이션 완료시 이벤트 - AnimationCompletedEvent

        /// <summary>
        /// 애니메이션 완료시 이벤트
        /// </summary>
        public static readonly RoutedEvent AnimationCompletedEvent = EventManager.RegisterRoutedEvent
        (
            "AnimationCompleted",
            RoutingStrategy.Bubble,
            typeof(AnimationCompletedEventHandler),
            typeof(AnimationBehavior)
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Dependency Property
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 소스 URI 속성 - SourceURIProperty

        /// <summary>
        /// 소스 URI 속성
        /// </summary>
        public static readonly DependencyProperty SourceURIProperty = DependencyProperty.RegisterAttached
        (
            "SourceURI",
            typeof(Uri),
            typeof(AnimationBehavior),
            new PropertyMetadata
            (
                null,
                SourcePropertyChangedCallback
            )
        );

        #endregion
        #region 소스 스트림 속성 - SourceStreamProperty

        /// <summary>
        /// 소스 스트림 속성
        /// </summary>
        public static readonly DependencyProperty SourceStreamProperty = DependencyProperty.RegisterAttached
        (
            "SourceStream",
            typeof(Stream),
            typeof(AnimationBehavior),
            new PropertyMetadata
            (
                null,
                SourcePropertyChangedCallback
            )
        );

        #endregion
        #region 반복 동작 속성 - RepeatBehaviorProperty

        /// <summary>
        /// 반복 동작 속성
        /// </summary>
        public static readonly DependencyProperty RepeatBehaviorProperty = DependencyProperty.RegisterAttached
        (
            "RepeatBehavior",
            typeof(RepeatBehavior),
            typeof(AnimationBehavior),
            new PropertyMetadata
            (
                default(RepeatBehavior),
                RepeatBehaviorPropertyChangedCallback
            )
        );

        #endregion
        #region 자동 시작 여부 속성 - AutoStartProperty

        /// <summary>
        /// 자동 시작 여부 속성
        /// </summary>
        public static readonly DependencyProperty AutoStartProperty = DependencyProperty.RegisterAttached
        (
            "AutoStart",
            typeof(bool),
            typeof(AnimationBehavior),
            new PropertyMetadata(true)
        );

        #endregion
        #region 디자인 모드에서 애니메이션 여부 속성 - AnimateInDesignModeProperty

        /// <summary>
        /// 디자인 모드에서 애니메이션 여부 속성
        /// </summary>
        public static readonly DependencyProperty AnimateInDesignModeProperty = DependencyProperty.RegisterAttached
        (
            "AnimateInDesignMode",
            typeof(bool),
            typeof(AnimationBehavior),
            new PropertyMetadata
            (
                false,
                AnimateInDesignModePropertyChangedCallback
            )
        );

        #endregion
        #region 애니메이터 속성 - AnimatorProperty

        /// <summary>
        /// 애니메이터 속성
        /// </summary>
        public static readonly DependencyProperty AnimatorProperty = DependencyProperty.RegisterAttached
        (
            "Animator",
            typeof(Animator),
            typeof(AnimationBehavior),
            new PropertyMetadata(null)
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region 시퀀스 번호 속성 - SequenceNumberProperty

        /// <summary>
        /// 시퀀스 번호 속성
        /// </summary>
        private static readonly DependencyProperty SequenceNumberProperty = DependencyProperty.RegisterAttached
        (
            "SequenceNumber",
            typeof(int),
            typeof(AnimationBehavior),
            new PropertyMetadata(0)
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 소스 URI 구하기 - GetSourceURI(image)

        /// <summary>
        /// 소스 URI 구하기
        /// </summary>
        /// <param name="image">이미지</param>
        /// <returns>URI</returns>
        [AttachedPropertyBrowsableForType(typeof(Image))]
        public static Uri GetSourceURI(Image image)
        {
            return (Uri)image.GetValue(SourceURIProperty);
        }

        #endregion
        #region 소스 URI 설정하기 - SetSourceURI(image, uri)

        /// <summary>
        /// 소스 URI 설정하기
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="uri">URI</param>
        public static void SetSourceURI(Image image, Uri uri)
        {
            image.SetValue(SourceURIProperty, uri);
        }

        #endregion

        #region 소스 스트림 구하기 - GetSourceStream(d)

        /// <summary>
        /// 소스 스트림 구하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <returns>스트림</returns>
        [AttachedPropertyBrowsableForType(typeof(Image))]
        public static Stream GetSourceStream(DependencyObject d)
        {
            return (Stream)d.GetValue(SourceStreamProperty);
        }

        #endregion
        #region 소스 스트림 설정하기 - SetSourceStream(d, stream)

        /// <summary>
        /// 소스 스트림 설정하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="stream">스트림</param>
        public static void SetSourceStream(DependencyObject d, Stream stream)
        {
            d.SetValue(SourceStreamProperty, stream);
        }

        #endregion

        #region 반복 동작 구하기 - GetRepeatBehavior(d)

        /// <summary>
        /// 반복 동작 구하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <returns>반복 동작</returns>
        [AttachedPropertyBrowsableForType(typeof(Image))]
        public static RepeatBehavior GetRepeatBehavior(DependencyObject d)
        {
            return (RepeatBehavior)d.GetValue(RepeatBehaviorProperty);
        }

        #endregion
        #region 반복 동작 설정하기 - SetRepeatBehavior(d, repeatBehavior)

        /// <summary>
        /// 반복 동작 설정하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="repeatBehavior">반복 동작</param>
        public static void SetRepeatBehavior(DependencyObject d, RepeatBehavior repeatBehavior)
        {
            d.SetValue(RepeatBehaviorProperty, repeatBehavior);
        }

        #endregion

        #region 자동 시작 여부 구하기 - GetAutoStart(d)

        /// <summary>
        /// 자동 시작 여부 구하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <returns>자동 시작 여부</returns>
        [AttachedPropertyBrowsableForType(typeof(Image))]
        public static bool GetAutoStart(DependencyObject d)
        {
            return (bool)d.GetValue(AutoStartProperty);
        }

        #endregion
        #region 자동 시작 여부 설정하기 - SetAutoStart(d, value)

        /// <summary>
        /// 자동 시작 여부 설정하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        public static void SetAutoStart(DependencyObject d, bool value)
        {
            d.SetValue(AutoStartProperty, value);
        }

        #endregion

        #region 디자인 모드에서 애니메이션 여부 구하기 - GetAnimateInDesignMode(d)

        /// <summary>
        /// 디자인 모드에서 애니메이션 여부 구하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <returns>디자인 모드에서 애니메이션 여부</returns>
        public static bool GetAnimateInDesignMode(DependencyObject d)
        {
            return (bool)d.GetValue(AnimateInDesignModeProperty);
        }

        #endregion
        #region 디자인 모드에서 애니메이션 여부 설정하기 - SetAnimateInDesignMode(d, value)

        /// <summary>
        /// 디자인 모드에서 애니메이션 여부 설정하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        public static void SetAnimateInDesignMode(DependencyObject d, bool value)
        {
            d.SetValue(AnimateInDesignModeProperty, value);
        }

        #endregion

        #region 애니메이터 구하기 - GetAnimator(d)

        /// <summary>
        /// 애니메이터 구하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <returns>애니메이터</returns>
        public static Animator GetAnimator(DependencyObject d)
        {
            return (Animator) d.GetValue(AnimatorProperty);
        }

        #endregion
        #region 애니메이터 설정하기 - SetAnimator(d, animator)

        /// <summary>
        /// 애니메이터 설정하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="animator">애니메이터</param>
        private static void SetAnimator(DependencyObject d, Animator animator)
        {
            d.SetValue(AnimatorProperty, animator);
        }

        #endregion

        #region 에러 핸들러 추가하기 - AddErrorHandler(d, handler)

        /// <summary>
        /// 에러 핸들러 추가하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="handler">핸들러</param>
        public static void AddErrorHandler(DependencyObject d, AnimationErrorEventHandler handler)
        {
            (d as UIElement)?.AddHandler(ErrorEvent, handler);
        }

        #endregion
        #region 에러 핸들러 제거하기 - RemoveErrorHandler(d, handler)

        /// <summary>
        /// 에러 핸들러 제거하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="handler">핸들러</param>
        public static void RemoveErrorHandler(DependencyObject d, AnimationErrorEventHandler handler)
        {
            (d as UIElement)?.RemoveHandler(ErrorEvent, handler);
        }

        #endregion

        #region 에러 이벤트 발생시키기 - FireErrorEvent(image, exception, errorKind)

        /// <summary>
        /// 에러 이벤트 발생시키기
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="exception">예외</param>
        /// <param name="errorKind">애니메이션 에러 종류</param>
        public static void FireErrorEvent(Image image, Exception exception, AnimationErrorKind errorKind)
        {
            image.RaiseEvent(new AnimationErrorEventArgs(image, exception, errorKind));
        }

        #endregion

        #region 다운로드 진행 핸들러 추가하기 - AddDownloadProgressHandler(d, handler)

        /// <summary>
        /// 다운로드 진행 핸들러 추가하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="handler">핸들러</param>
        public static void AddDownloadProgressHandler(DependencyObject d, DownloadProgressEventHandler handler)
        {
            (d as UIElement)?.AddHandler(DownloadProgressEvent, handler);
        }

        #endregion
        #region 다운로드 진행 핸들러 제거하기 - RemoveDownloadProgressHandler(d, handler)

        /// <summary>
        /// 다운로드 진행 핸들러 제거하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="handler">핸들러</param>
        public static void RemoveDownloadProgressHandler(DependencyObject d, DownloadProgressEventHandler handler)
        {
            (d as UIElement)?.RemoveHandler(DownloadProgressEvent, handler);
        }

        #endregion

        #region 다운로드 진행 이벤트 발생시키기 - FireDownloadProgressEvent(image, progressRatio)

        /// <summary>
        /// 다운로드 진행 이벤트 발생시키기
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="progressRatio">진행률</param>
        public static void FireDownloadProgressEvent(Image image, int progressRatio)
        {
            image.RaiseEvent(new DownloadProgressEventArgs(image, progressRatio));
        }

        #endregion

        #region 로드시 핸들러 추가하기 - AddLoadedHandler(d, handler)

        /// <summary>
        /// 로드시 핸들러 추가하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="handler">핸들러</param>
        public static void AddLoadedHandler(DependencyObject d, RoutedEventHandler handler)
        {
            (d as UIElement)?.AddHandler(LoadedEvent, handler);
        }

        #endregion
        #region 로드시 핸들러 제거하기 - RemoveLoadedHandler(d, handler)

        /// <summary>
        /// 로드시 핸들러 제거하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="handler">핸들러</param>
        public static void RemoveLoadedHandler(DependencyObject d, RoutedEventHandler handler)
        {
            (d as UIElement)?.RemoveHandler(LoadedEvent, handler);
        }

        #endregion

        #region 로드시 이벤트 발생시키기 - FireLoadedEvent(image)

        /// <summary>
        /// 로드시 이벤트 발생시키기
        /// </summary>
        /// <param name="image">이미지</param>
        public static void FireLoadedEvent(Image image)
        {
            image.RaiseEvent(new RoutedEventArgs(LoadedEvent, image));
        }

        #endregion

        #region 애니메이션 완료시 핸들러 추가하기 - AddAnimationCompletedHandler(d, handler)

        /// <summary>
        /// 애니메이션 완료시 핸들러 추가하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="handler">핸들러</param>
        public static void AddAnimationCompletedHandler(DependencyObject d, AnimationCompletedEventHandler handler)
        {
            (d as UIElement)?.AddHandler(AnimationCompletedEvent, handler);
        }

        #endregion
        #region 애니메이션 완료시 핸들러 제거하기 - RemoveAnimationCompletedHandler(d, handler)

        /// <summary>
        /// 애니메이션 완료시 핸들러 제거하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="handler">핸들러</param>
        public static void RemoveAnimationCompletedHandler(DependencyObject d, AnimationCompletedEventHandler handler)
        {
            (d as UIElement)?.RemoveHandler(AnimationCompletedEvent, handler);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private
        ////////////////////////////////////////////////////////////////////// Callback

        #region 소스 속성 변경시 콜백 처리하기 - SourcePropertyChangedCallback(d, e)

        /// <summary>
        /// 소스 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void SourcePropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if(d is not Image image)
            {
                return;
            }

            InitializeAnimation(image);
        }

        #endregion
        #region 반복 동작 속성 변경시 콜백 처리하기 - RepeatBehaviorPropertyChangedCallback(d, e)

        /// <summary>
        /// 반복 동작 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void RepeatBehaviorPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            GetAnimator(d)?.OnRepeatBehaviorChanged();
        }

        #endregion
        #region 디자인 모드에서 애니메이션 여부 속성 변경시 콜백 처리하기 - AnimateInDesignModePropertyChangedCallback(d, e)

        /// <summary>
        /// 디자인 모드에서 애니메이션 여부 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void AnimateInDesignModePropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if(d is not Image image)
            {
                return;
            }

            InitializeAnimation(image);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////// Attached Method

        #region 시퀀스 번호 구하기 - GetSequenceNumber(d)

        /// <summary>
        /// 시퀀스 번호 구하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <returns>시퀀스 번호</returns>
        private static int GetSequenceNumber(DependencyObject d)
        {
            return (int)d.GetValue(SequenceNumberProperty);
        }

        #endregion
        #region 시퀀스 번호 설정하기 - SetSequenceNumber(d, sequenceNumber)

        /// <summary>
        /// 시퀀스 번호 설정하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="sequenceNumber">시퀀스 번호</param>
        private static void SetSequenceNumber(DependencyObject d, int sequenceNumber)
        {
            d.SetValue(SequenceNumberProperty, sequenceNumber);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////// Event

        #region 이미지 로드시 처리하기 - image_Loaded(sender, e)

        /// <summary>
        /// 이미지 로드시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private static void image_Loaded(object sender, RoutedEventArgs e)
        {
            Image image = sender as Image;

            image.Loaded -= image_Loaded;

            InitializeAnimation(image);
        }

        #endregion
        #region 이미지 언로드시 처리하기 - image_Unloaded(sender, e)

        /// <summary>
        /// 이미지 언로드시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private static void image_Unloaded(object sender, RoutedEventArgs e)
        {
            Image image = sender as Image;

            image.Unloaded -= image_Unloaded;
            image.Loaded   += image_Loaded;

            int sequenceNumber = GetSequenceNumber(image) + 1;

            SetSequenceNumber(image, sequenceNumber);

            image.Source = null;

            ClearAnimatorCore(image);
        }

        #endregion

        #region 애니메이터 에러시 처리하기 - animator_Error(sender, e)

        /// <summary>
        /// 애니메이터 에러시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private static void animator_Error(object sender, AnimationErrorEventArgs e)
        {
            UIElement uiElement = e.Source as UIElement;

            uiElement?.RaiseEvent(e);
        }

        #endregion
        #region 애니메이터 애니메이션 완료시 처리하기 - animator_AnimationCompleted(sender, e)

        /// <summary>
        /// 애니메이터 애니메이션 완료시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private static void animator_AnimationCompleted(object sender, AnimationCompletedEventArgs e)
        {
            Image image = e.Source as Image;

            image?.RaiseEvent(e);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////// Function

        #region 로드 여부 구하기 - IsLoaded(frameworkElement)

        /// <summary>
        /// 로드 여부 구하기
        /// </summary>
        /// <param name="frameworkElement">프레임워크 엘리먼트</param>
        /// <returns>로드 여부</returns>
        private static bool IsLoaded(FrameworkElement frameworkElement)
        {
            return frameworkElement.IsLoaded;
        }

        #endregion
        #region 애니메이터 지우기 (코어) - ClearAnimatorCore(image)

        /// <summary>
        /// 애니메이터 지우기 (코어)
        /// </summary>
        /// <param name="image">이미지</param>
        private static void ClearAnimatorCore(Image image)
        {
            Animator animator = GetAnimator(image);

            if(animator == null)
            {
                return;
            }

            animator.AnimationCompleted -= animator_AnimationCompleted;
            animator.Error              -= animator_Error;

            animator.Dispose();

            SetAnimator(image, null);
        }

        #endregion
        #region 디자인 모드 여부 구하기 - IsInDesignMode(d)

        /// <summary>
        /// 디자인 모드 여부 구하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <returns>디자인 모드 여부</returns>
        private static bool IsInDesignMode(DependencyObject d)
        {
            return DesignerProperties.GetIsInDesignMode(d);
        }

        #endregion
        #region 정적 이미지 설정하기 (코어) - SetStaticImageCore(image, stream)

        /// <summary>
        /// 정적 이미지 설정하기 (코어)
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="stream">스트림</param>
        private static void SetStaticImageCore(Image image, Stream stream)
        {
            stream.Seek(0, SeekOrigin.Begin);

            BitmapImage bitmapImage = new BitmapImage();

            bitmapImage.BeginInit();

            bitmapImage.StreamSource = stream;

            bitmapImage.EndInit();

            image.Source = bitmapImage;
        }

        #endregion
        #region 정적 이미지 설정하기 - SetStaticImage(image, stream)

        /// <summary>
        /// 정적 이미지 설정하기
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="stream">스트림</param>
        private static void SetStaticImage(Image image, Stream stream)
        {
            try
            {
                SetStaticImageCore(image, stream);
            }
            catch(Exception exception)
            {
                FireErrorEvent(image, exception, AnimationErrorKind.Loading);
            }
        }

        #endregion
        #region 디자인 모드 조사하기 - CheckDesignMode(image, sourceURI, sourceStream)

        /// <summary>
        /// 디자인 모드 조사하기
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="sourceURI">소스 URI</param>
        /// <param name="sourceStream">소스 스트림</param>
        /// <returns>디자인 모드 조사 결과</returns>
        private static bool CheckDesignMode(Image image, Uri sourceURI, Stream sourceStream)
        {
            if(IsInDesignMode(image) && !GetAnimateInDesignMode(image))
            {
                try
                {
                    if(sourceStream != null)
                    {
                        SetStaticImage(image, sourceStream);
                    }
                    else if(sourceURI != null)
                    {
                        BitmapImage bitmapImage = new BitmapImage
                        {
                            UriSource = sourceURI
                        };

                        image.Source = bitmapImage;
                    }
                }
                catch
                {
                    image.Source = null;
                }

                return false;
            }

            return true;
        }

        #endregion
        #region 애니메이션 설정하기 (코어, 비동기) - SetAnimatorCoreAsync(image, animator)

        /// <summary>
        /// 애니메이션 설정하기 (코어, 비동기)
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="animator">애니메이터</param>
        /// <returns>태스크</returns>
        private static async Task SetAnimatorCoreAsync(Image image, Animator animator)
        {
            SetAnimator(image, animator);

            animator.Error              += animator_Error;
            animator.AnimationCompleted += animator_AnimationCompleted;

            image.Source = animator.Bitmap;

            if(GetAutoStart(image))
            {
                animator.Play();
            }
            else
            {
                await animator.ShowFirstFrameAsync();
            }
        }

        #endregion
        #region 애니메이션 초기화하기 (비동기) - InitializeAnimationAsync(image, stream, repeatBehavior, sequenceNumber)

        /// <summary>
        /// 애니메이션 초기화하기 (비동기)
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="stream">스트림</param>
        /// <param name="repeatBehavior">반복 동작</param>
        /// <param name="sequenceNumber">시퀀스 번호</param>
        private static async void InitializeAnimationAsync(Image image, Stream stream, RepeatBehavior repeatBehavior, int sequenceNumber)
        {
            if(!CheckDesignMode(image, null, stream))
            {
                return;
            }

            try
            {
                Animator animator = await ImageAnimator.CreateAsync(stream, repeatBehavior, image);

                if(GetSequenceNumber(image) != sequenceNumber)
                {
                    animator.Dispose();

                    return;
                }

                await SetAnimatorCoreAsync(image, animator);

                FireLoadedEvent(image);
            }
            catch(InvalidSignatureException)
            {
                SetStaticImage(image, stream);

                FireLoadedEvent(image);
            }
            catch(Exception exception)
            {
                FireErrorEvent(image, exception, AnimationErrorKind.Loading);
            }
        }

        #endregion
        #region 절대 URI 구하기 - GetAbsoluteURI(image)

        /// <summary>
        /// 절대 URI 구하기
        /// </summary>
        /// <param name="image">이미지</param>
        /// <returns>URI</returns>
        private static Uri GetAbsoluteURI(Image image)
        {
            Uri uri = GetSourceURI(image);

            if(uri == null)
            {
                return null;
            }

            if(!uri.IsAbsoluteUri)
            {
                Uri baseUri = ((IUriContext)image).BaseUri;

                if(baseUri != null)
                {
                    uri = new Uri(baseUri, uri);
                }
                else
                {
                    throw new InvalidOperationException("Relative URI can't be resolved");
                }
            }

            return uri;
        }

        #endregion
        #region 정적 이미지 설정하기 (비동기) - SetStaticImageAsync(image, sourceURI)

        /// <summary>
        /// 정적 이미지 설정하기 (비동기)
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="sourceURI">소스 URI</param>
        /// <returns></returns>
        private static async Task SetStaticImageAsync(Image image, Uri sourceURI)
        {
            try
            {
                Progress<int> progress = new Progress<int>(percentage => FireDownloadProgressEvent(image, percentage));

                Stream stream = await URILoader.GetStreamFromURIAsync(sourceURI, progress);

                SetStaticImageCore(image, stream);
            }
            catch(Exception exception)
            {
                FireErrorEvent(image, exception, AnimationErrorKind.Loading);
            }
        }

        #endregion
        #region 애니메이션 초기화하기 (비동기) - InitializeAnimationAsync(image, sourceURI, repeatBehavior, sequenceNumber)

        /// <summary>
        /// 애니메이션 초기화하기 (비동기)
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="sourceURI">소스 URI</param>
        /// <param name="repeatBehavior">반복 동작</param>
        /// <param name="sequenceNumber">시퀀스 번호</param>
        private static async void InitializeAnimationAsync
        (
            Image          image,
            Uri            sourceURI,
            RepeatBehavior repeatBehavior,
            int            sequenceNumber
        )
        {
            if(!CheckDesignMode(image, sourceURI, null))
            {
                return;
            }

            try
            {
                Progress<int> progress = new Progress<int>(percentage => FireDownloadProgressEvent(image, percentage));

                ImageAnimator animator = await ImageAnimator.CreateAsync(sourceURI, repeatBehavior, progress, image);

                if(GetSequenceNumber(image) != sequenceNumber)
                {
                    animator.Dispose();

                    return;
                }

                await SetAnimatorCoreAsync(image, animator);

                FireLoadedEvent(image);
            }
            catch(InvalidSignatureException)
            {
                await SetStaticImageAsync(image, sourceURI);

                FireLoadedEvent(image);
            }
            catch(Exception exception)
            {
                FireErrorEvent(image, exception, AnimationErrorKind.Loading);
            }
        }

        #endregion
        #region 애니메이션 초기화하기 - InitializeAnimation(image)

        /// <summary>
        /// 애니메이션 초기화하기
        /// </summary>
        /// <param name="image">이미지</param>
        private static void InitializeAnimation(Image image)
        {
            if(IsLoaded(image))
            {
                image.Unloaded += image_Unloaded;
            }
            else
            {
                image.Loaded += image_Loaded;

                return;
            }

            int sequenceNumber = GetSequenceNumber(image) + 1;

            SetSequenceNumber(image, sequenceNumber);

            image.Source = null;

            ClearAnimatorCore(image);

            try
            {
                Stream stream = GetSourceStream(image);

                if(stream != null)
                {
                    InitializeAnimationAsync
                    (
                        image,
                        stream.AsBuffered(),
                        GetRepeatBehavior(image),
                        sequenceNumber
                    );

                    return;
                }

                Uri uri = GetAbsoluteURI(image);

                if(uri != null)
                {
                    InitializeAnimationAsync(image, uri, GetRepeatBehavior(image), sequenceNumber);
                }
            }
            catch(Exception exception)
            {
                FireErrorEvent(image, exception, AnimationErrorKind.Loading);
            }
        }

        #endregion
    }
}