﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Media.Imaging;
using System.Runtime.InteropServices;

using TestLibrary.Decoding;
using TestLibrary.Decompression;
using TestLibrary.Extension;

namespace TestLibrary
{
    /// <summary>
    /// 애니메이터
    /// </summary>
    public abstract class Animator : DependencyObject, IDisposable
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Class
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region GIF 팔레트 - GIFPalette

        /// <summary>
        /// GIF 팔레트
        /// </summary>
        private class GIFPalette
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Field
            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region Field

            /// <summary>
            /// 색상 배열
            /// </summary>
            private readonly Color[] colorArray;

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 투명 인덱스 - TransparencyIndex

            /// <summary>
            /// 투명 인덱스
            /// </summary>
            public int? TransparencyIndex { get; }

            #endregion
            #region 인덱서 - this[index]

            /// <summary>
            /// 인덱서
            /// </summary>
            /// <param name="index">인덱스</param>
            /// <returns>색상</returns>
            public Color this[int index] => this.colorArray[index];

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - GIFPalette(transparencyIndex, colorArray)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="transparencyIndex">투명 인덱스</param>
            /// <param name="colorArray">색상 배열</param>
            public GIFPalette(int? transparencyIndex, Color[] colorArray)
            {
                TransparencyIndex = transparencyIndex;

                this.colorArray = colorArray;
            }

            #endregion
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 에러 이벤트 - Error

        /// <summary>
        /// 에러 이벤트
        /// </summary>
        public event EventHandler<AnimationErrorEventArgs> Error;

        #endregion
        #region 현재 프레임 변경시 이벤트 - CurrentFrameChanged

        /// <summary>
        /// 현재 프레임 변경시 이벤트
        /// </summary>
        public event EventHandler CurrentFrameChanged;

        #endregion
        #region 애니메이션 완료시 이벤트 - AnimationCompleted

        /// <summary>
        /// 애니메이션 완료시 이벤트
        /// </summary>
        public event EventHandler<AnimationCompletedEventArgs> AnimationCompleted;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 소스 스트림
        /// </summary>
        private readonly Stream sourceStream;

        /// <summary>
        /// 소스 URI
        /// </summary>
        private readonly Uri sourceURI;

        /// <summary>
        /// 소스 스트림 소유자 여부
        /// </summary>
        private readonly bool isSourceStreamOwner;

        /// <summary>
        /// GIF 데이터 스트림
        /// </summary>
        private readonly GIFDataStream dataStream;

        /// <summary>
        /// GIF 팔레트 딕셔너리
        /// </summary>
        private readonly Dictionary<int, GIFPalette> paletteDictionary;

        /// <summary>
        /// 쓰기 가능 비트맵
        /// </summary>
        private readonly WriteableBitmap bitmap;

        /// <summary>
        /// 스트라이드
        /// </summary>
        private readonly int stride;

        /// <summary>
        /// 이전 백 버퍼
        /// </summary>
        private readonly byte[] previousBackBuffer;

        /// <summary>
        /// 인덱스 스트림 버퍼
        /// </summary>
        private readonly byte[] indexStreamBuffer;

        /// <summary>
        /// 타이밍 관리자
        /// </summary>
        private readonly TimingManager timingManager;

        /// <summary>
        /// 시작 여부
        /// </summary>
        private bool isStarted;

        /// <summary>
        /// 취소 토큰 소스
        /// </summary>
        private CancellationTokenSource cancellationTokenSource;

        /// <summary>
        /// 프레임 인덱스
        /// </summary>
        private int frameIndex;

        /// <summary>
        /// 이전 프레임 인덱스
        /// </summary>
        private int previousFrameIndex;

        /// <summary>
        /// 이전 GIF 프레임
        /// </summary>
        private GIFFrame previousFrame;

        /// <summary>
        /// 처분 예정 여부
        /// </summary>
        private volatile bool disposing;

        /// <summary>
        /// 처분 여부
        /// </summary>
        private bool disposed;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 프레임 카운트 - FrameCount

        /// <summary>
        /// 프레임 카운트
        /// </summary>
        public int FrameCount => this.dataStream.FrameList.Count;

        #endregion
        #region 중지 여부 - IsPaused

        /// <summary>
        /// 중지 여부
        /// </summary>
        public bool IsPaused => this.timingManager.IsPaused;

        #endregion
        #region 완료 여부 - IsCompleted

        /// <summary>
        /// 완료 여부
        /// </summary>
        public bool IsCompleted
        {
            get
            {
                if(this.isStarted)
                {
                    return this.timingManager.IsComplete;
                }

                return false;
            }
        }

        #endregion
        #region 현재 프레임 인덱스 - CurrentFrameIndex

        /// <summary>
        /// 현재 프레임 인덱스
        /// </summary>
        public int CurrentFrameIndex
        {
            get
            {
                return this.frameIndex;
            }
            internal set
            {
                this.frameIndex = value;

                OnCurrentFrameChanged();
            }
        }

        #endregion
        #region 비트맵 - Bitmap

        /// <summary>
        /// 비트맵
        /// </summary>
        public BitmapSource Bitmap => this.bitmap;

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Proected

        #region 애니메이션 소스 - AnimationSource

        /// <summary>
        /// 애니메이션 소스
        /// </summary>
        protected abstract object AnimationSource { get; }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - Animator(sourceStream, sourceURI, dataStream, repeatBehavior)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="sourceStream">소스 스트림</param>
        /// <param name="sourceURI">소스 URI</param>
        /// <param name="dataStream">GIF 데이터 스트림</param>
        /// <param name="repeatBehavior">반복 동작</param>
        public Animator(Stream sourceStream, Uri sourceURI, GIFDataStream dataStream, RepeatBehavior repeatBehavior)
        {
            this.sourceStream        = sourceStream;
            this.sourceURI           = sourceURI;
            this.isSourceStreamOwner = sourceURI != null;
            this.dataStream          = dataStream;
            this.paletteDictionary   = GetPaletteDictionary(dataStream);
            this.bitmap              = GetBitmap(dataStream);

            GIFLogicalScreenDescriptor descriptor = dataStream.Header.LogicalScreenDescriptor;

            this.stride              = 4 * ((descriptor.Width * 32 + 31) / 32);
            this.previousBackBuffer  = new byte[descriptor.Height * this.stride];
            this.indexStreamBuffer   = GetIndexStreamBuffer(dataStream, this.sourceStream);
            this.timingManager       = GetTimingManager(dataStream, repeatBehavior);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Descturctor

        #region 소멸자 - ~Animator()

        /// <summary>
        /// 소멸자
        /// </summary>
        ~Animator()
        {
            Dispose(false);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 생성하기 (코어, 비동기) - CreateCoreAsync<TAnimator>(sourceURI, progress, createFunc)

        /// <summary>
        /// 생성하기 (코어, 비동기)
        /// </summary>
        /// <typeparam name="TAnimator">애니메이터 타입</typeparam>
        /// <param name="sourceURI">소스 URI</param>
        /// <param name="progress">진행 인터페이스</param>
        /// <param name="createFunc">생성하기 함수</param>
        /// <returns>애니메이터 태스크</returns>
        public static async Task<TAnimator> CreateCoreAsync<TAnimator>
        (
            Uri                                    sourceURI,
            IProgress<int>                         progress,
            Func<Stream, GIFDataStream, TAnimator> createFunc
        )
        where TAnimator : Animator
        {
            Stream stream = await URILoader.GetStreamFromURIAsync(sourceURI, progress);

            try
            {
                return await CreateCoreAsync(stream, metadata => createFunc(stream, metadata));
            }
            catch
            {
                stream?.Dispose();

                throw;
            }
        }

        #endregion
        #region 생성하기 (코어, 비동기) - CreateCoreAsync<TAnimator>(sourceStream, createFunc)

        /// <summary>
        /// 생성하기 (코어, 비동기)
        /// </summary>
        /// <typeparam name="TAnimator">애니메이터 타입</typeparam>
        /// <param name="sourceStream">소스 스트림</param>
        /// <param name="createFunc">생성하기 함수</param>
        /// <returns>애니메이터 태스크</returns>
        public static async Task<TAnimator> CreateCoreAsync<TAnimator>(Stream sourceStream, Func<GIFDataStream, TAnimator> createFunc) where TAnimator : Animator
        {
            if(!sourceStream.CanSeek)
            {
                throw new ArgumentException("The stream is not seekable");
            }

            sourceStream.Seek(0, SeekOrigin.Begin);

            GIFDataStream dataStream = await GIFDataStream.ReadAsync(sourceStream);

            return createFunc(dataStream);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 첫번째 프레임 표시하기 (비동기) - ShowFirstFrameAsync()

        /// <summary>
        /// 첫번째 프레임 표시하기 (비동기)
        /// </summary>
        /// <returns>태스크</returns>
        public async Task ShowFirstFrameAsync()
        {
            try
            {
                await RenderFrameAsync(0, CancellationToken.None);

                CurrentFrameIndex = 0;

                this.timingManager.Pause();
            }
            catch(Exception exception)
            {
                OnError(exception, AnimationErrorKind.Rendering);
            }
        }

        #endregion
        #region 재생하기 - Play()

        /// <summary>
        /// 재생하기
        /// </summary>
        public async void Play()
        {
            try
            {
                if(this.timingManager.IsComplete)
                {
                    this.timingManager.Reset();

                    this.isStarted = false;
                }

                if(!this.isStarted)
                {
                    this.cancellationTokenSource?.Dispose();

                    this.cancellationTokenSource = new CancellationTokenSource();

                    this.isStarted = true;

                    if(this.timingManager.IsPaused)
                    {
                        this.timingManager.Resume();
                    }

                    await RunAsync(this.cancellationTokenSource.Token);
                }
                else if(this.timingManager.IsPaused)
                {
                    this.timingManager.Resume();
                }
            }
            catch(OperationCanceledException)
            {
            }
            catch(Exception exception)
            {
                if(!this.disposing)
                {
                    OnError(exception, AnimationErrorKind.Rendering);
                }
            }
        }

        #endregion
        #region 중지하기 - Pause()

        /// <summary>
        /// 중지하기
        /// </summary>
        public void Pause()
        {
            this.timingManager.Pause();
        }

        #endregion
        #region 되감기 - Rewind()

        /// <summary>
        /// 되감기
        /// </summary>
        public async void Rewind()
        {
            CurrentFrameIndex = 0;

            bool isStopped = this.timingManager.IsPaused || this.timingManager.IsComplete;

            this.timingManager.Reset();

            if(isStopped)
            {
                this.timingManager.Pause();

                this.isStarted = false;

                try
                {
                    await RenderFrameAsync(0, CancellationToken.None);
                }
                catch(Exception exception)
                {
                    OnError(exception, AnimationErrorKind.Rendering);
                }
            }
        }

        #endregion

        #region 문자열 구하기 - ToString()

        /// <summary>
        /// 문자열 구하기
        /// </summary>
        /// <returns>문자열</returns>
        public override string ToString()
        {
            string text = this.sourceURI?.ToString() ?? this.sourceStream.ToString();

            return "GIF: " + text;
        }

        #endregion
        #region 처분하기 - Dispose()

        /// <summary>
        /// 처분하기
        /// </summary>
        public void Dispose()
        {
            Dispose(true);

            GC.SuppressFinalize(this);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Protected

        #region 에러시 처리하기 - OnError(exception, errorKind)

        /// <summary>
        /// 에러시 처리하기
        /// </summary>
        /// <param name="exception">예외</param>
        /// <param name="errorKind">애니메이션 에러 종류</param>
        protected virtual void OnError(Exception exception, AnimationErrorKind errorKind)
        {
            Error?.Invoke(this, new AnimationErrorEventArgs(AnimationSource, exception, errorKind));
        }

        #endregion
        #region 현재 프레임 변경시 처리하기 - OnCurrentFrameChanged()

        /// <summary>
        /// 현재 프레임 변경시 처리하기
        /// </summary>
        protected virtual void OnCurrentFrameChanged()
        {
            CurrentFrameChanged?.Invoke(this, EventArgs.Empty);
        }

        #endregion
        #region 반복 동작 변경시 처리하기 - OnRepeatBehaviorChanged()

        /// <summary>
        /// 반복 동작 변경시 처리하기
        /// </summary>
        public void OnRepeatBehaviorChanged()
        {
            if(this.timingManager == null)
            {
                return;
            }

            RepeatBehavior newRepeatBehavior = GetSpecifiedRepeatBehavior();

            RepeatBehavior newActualRepeatBehavior = GetActualRepeatBehavior(this.dataStream, newRepeatBehavior);

            if(this.timingManager.RepeatBehavior == newActualRepeatBehavior)
            {
                return;
            }

            this.timingManager.RepeatBehavior = newActualRepeatBehavior;

            Rewind();
        }

        #endregion
        #region 애니메이션 완료시 처리하기 - OnAnimationCompleted()

        /// <summary>
        /// 애니메이션 완료시 처리하기
        /// </summary>
        protected virtual void OnAnimationCompleted()
        {
            AnimationCompleted?.Invoke(this, new AnimationCompletedEventArgs(AnimationSource));
        }

        #endregion

        #region 지정 반복 동작 구하기 - GetSpecifiedRepeatBehavior()

        /// <summary>
        /// 지정 반복 동작 구하기
        /// </summary>
        /// <returns>반복 동작</returns>
        protected abstract RepeatBehavior GetSpecifiedRepeatBehavior();

        #endregion

        #region 처분하기 - Dispose(disposing)

        /// <summary>
        /// 처분하기
        /// </summary>
        /// <param name="disposing">처분 예정 여부</param>
        protected virtual void Dispose(bool disposing)
        {
            if(!this.disposed)
            {
                this.disposing = true;

                if(this.timingManager != null)
                {
                    this.timingManager.Completed -= timingManager_Completed;
                }

                this.cancellationTokenSource?.Cancel();

                if(this.isSourceStreamOwner)
                {
                    try
                    {
                        this.sourceStream?.Dispose();
                    }
                    catch
                    {
                    }
                }

                this.disposed = true;
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 타이밍 관리자 완료시 처리하기 - timingManager_Completed(sender, e)

        /// <summary>
        /// 타이밍 관리자 완료시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void timingManager_Completed(object sender, EventArgs e)
        {
            OnAnimationCompleted();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region GIF 팔레트 딕셔너리 구하기 - GetPaletteDictionary(dataStream)

        /// <summary>
        /// GIF 팔레트 딕셔너리 구하기
        /// </summary>
        /// <param name="dataStream">GIF 데이터 스트림</param>
        /// <returns>GIF 팔레트 딕셔너리</returns>
        private Dictionary<int, GIFPalette> GetPaletteDictionary(GIFDataStream dataStream)
        {
            Dictionary<int, GIFPalette> paletteDictionary = new Dictionary<int, GIFPalette>();

            Color[] globalColorTable = null;

            if(dataStream.Header.LogicalScreenDescriptor.HasGlobalColorTable)
            {
                globalColorTable = dataStream.GlobalColorTable.Select(gc => Color.FromArgb(0xFF, gc.Red, gc.Green, gc.Blue)).ToArray();
            }

            for(int i = 0; i < dataStream.FrameList.Count; i++)
            {
                GIFFrame frame = dataStream.FrameList[i];

                Color[] colorTable = globalColorTable;

                if(frame.Descriptor.HasLocalColorTable)
                {
                    colorTable = frame.LocalColorTable.Select(gc => Color.FromArgb(0xFF, gc.Red, gc.Green, gc.Blue)).ToArray();
                }

                int? transparencyIndex = null;

                GIFGraphicControlExtension extension = frame.GraphicControl;

                if(extension != null && extension.HasTransparency)
                {
                    transparencyIndex = extension.TransparencyIndex;
                }

                paletteDictionary[i] = new GIFPalette(transparencyIndex, colorTable);
            }

            return paletteDictionary;
        }

        #endregion
        #region 쓰기 가능 비트맵 구하기 - GetBitmap(dataStream)

        /// <summary>
        /// 쓰기 가능 비트맵 구하기
        /// </summary>
        /// <param name="dataStream">데이터 스트림</param>
        /// <returns>쓰기 가능 비트맵</returns>
        private WriteableBitmap GetBitmap(GIFDataStream dataStream)
        {
            GIFLogicalScreenDescriptor descriptor = dataStream.Header.LogicalScreenDescriptor;

            WriteableBitmap bitmap = new WriteableBitmap(descriptor.Width, descriptor.Height, 96, 96, PixelFormats.Bgra32, null);

            return bitmap;
        }

        #endregion
        #region 인덱스 스트림 버퍼 구하기 - GetIndexStreamBuffer(dataStream, stream)

        /// <summary>
        /// 인덱스 스트림 버퍼 구하기
        /// </summary>
        /// <param name="dataStream">데이터 스트림</param>
        /// <param name="stream">스트림</param>
        /// <returns>인덱스 스트림 버퍼</returns>
        private static byte[] GetIndexStreamBuffer(GIFDataStream dataStream, Stream stream)
        {
            long lastSize    = stream.Length - dataStream.FrameList.Last().ImageData.CompressedDataStartOffset;
            long maximumSize = lastSize;

            if(dataStream.FrameList.Count > 1)
            {
                IEnumerable<long> sizeEnumerable = dataStream.FrameList.Zip
                (
                    dataStream.FrameList.Skip(1),
                    (f1, f2) => f2.ImageData.CompressedDataStartOffset - f1.ImageData.CompressedDataStartOffset
                );

                maximumSize = Math.Max(sizeEnumerable.Max(), lastSize);
            }

            return new byte[maximumSize + 4];
        }

        #endregion
        #region GIF에서 반복 동작 구하기 - GetRepeatBehaviorFromGIF(dataStream)

        /// <summary>
        /// GIF에서 반복 동작 구하기
        /// </summary>
        /// <param name="dataStream">GIF 데이터 스트림</param>
        /// <returns>반복 동작</returns>
        private RepeatBehavior GetRepeatBehaviorFromGIF(GIFDataStream dataStream)
        {
            if(dataStream.RepeatCount == 0)
            {
                return RepeatBehavior.Forever;
            }

            return new RepeatBehavior(dataStream.RepeatCount);
        }

        #endregion
        #region 실제 반복 동작 구하기 - GetActualRepeatBehavior(dataStream, repeatBehavior)

        /// <summary>
        /// 실제 반복 동작 구하기
        /// </summary>
        /// <param name="dataStream">데이터 스트림</param>
        /// <param name="repeatBehavior">반복 동작</param>
        /// <returns>반복 동작</returns>
        private RepeatBehavior GetActualRepeatBehavior(GIFDataStream dataStream, RepeatBehavior repeatBehavior)
        {
            return repeatBehavior == default ? GetRepeatBehaviorFromGIF(dataStream) : repeatBehavior;
        }

        #endregion
        #region 프레임 지연 구하기 - GetFrameDelay(frame)

        /// <summary>
        /// 프레임 지연 구하기
        /// </summary>
        /// <param name="frame">GIF 프레임</param>
        /// <returns>시간 범위</returns>
        private TimeSpan GetFrameDelay(GIFFrame frame)
        {
            GIFGraphicControlExtension extension = frame.GraphicControl;

            if(extension != null)
            {
                if(extension.Delay != 0)
                {
                    return TimeSpan.FromMilliseconds(extension.Delay);
                }
            }

            return TimeSpan.FromMilliseconds(100);
        }

        #endregion
        #region 타이밍 관리자 구하기 - GetTimingManager(dataStream, repeatBehavior)

        /// <summary>
        /// 타이밍 관리자 구하기
        /// </summary>
        /// <param name="dataStream">GIF 데이터 스트림</param>
        /// <param name="repeatBehavior">반복 동작</param>
        /// <returns>타이밍 관리자</returns>
        private TimingManager GetTimingManager(GIFDataStream dataStream, RepeatBehavior repeatBehavior)
        {
            RepeatBehavior actualRepeatBehavior = GetActualRepeatBehavior(dataStream, repeatBehavior);

            TimingManager timingManager = new TimingManager(actualRepeatBehavior);

            foreach(GIFFrame frame in dataStream.FrameList)
            {
                timingManager.Add(GetFrameDelay(frame));
            }

            timingManager.Completed += timingManager_Completed;

            return timingManager;
        }

        #endregion

        #region 고정 프레임 사각형 구하기 - GetFixedUpFrameRectangle(descriptor)

        /// <summary>
        /// 고정 프레임 사각형 구하기
        /// </summary>
        /// <param name="descriptor">GIF 이미지 설명자</param>
        /// <returns>정수 사각형</returns>
        private Int32Rect GetFixedUpFrameRectangle(GIFImageDescriptor descriptor)
        {
            int width  = Math.Min(descriptor.Width , this.bitmap.PixelWidth  - descriptor.Left);
            int height = Math.Min(descriptor.Height, this.bitmap.PixelHeight - descriptor.Top );

            return new Int32Rect(descriptor.Left, descriptor.Top, width, height);
        }

        #endregion
        #region 인덱스 스트림 구하기 (비동기) - GetIndexStreamAsync(frame, cancellationToken)

        /// <summary>
        /// 인덱스 스트림 구하기 (비동기)
        /// </summary>
        /// <param name="frame">프레임</param>
        /// <param name="cancellationToken">취소 토큰</param>
        /// <returns>스트림 태스크</returns>
        private async Task<Stream> GetIndexStreamAsync(GIFFrame frame, CancellationToken cancellationToken)
        {
            GIFImageData imageData = frame.ImageData;

            cancellationToken.ThrowIfCancellationRequested();

            this.sourceStream.Seek(imageData.CompressedDataStartOffset, SeekOrigin.Begin);

            using(MemoryStream memoryStream = new MemoryStream(this.indexStreamBuffer))
            {
                await GIFHelper.CopyDataBlocksToStreamAsync
                (
                    this.sourceStream,
                    memoryStream,
                    cancellationToken
                )
                .ConfigureAwait(false);
            }

            LZWDecompressStream stream = new LZWDecompressStream(this.indexStreamBuffer, imageData.LZWMinimumCodeSize);

            return stream;
        }

        #endregion
        #region 비트맵에 복사하기 - CopyToBitmap(sourceByteArry, targetBitmap, offset, length)

        /// <summary>
        /// 비트맵에 복사하기
        /// </summary>
        /// <param name="sourceByteArry">소스 바이트 배열</param>
        /// <param name="targetBitmap">타겟 쓰기 가능 비트맵</param>
        /// <param name="offset">오프셋</param>
        /// <param name="length">길이</param>
        private static void CopyToBitmap(byte[] sourceByteArry, WriteableBitmap targetBitmap, int offset, int length)
        {
            Marshal.Copy(sourceByteArry, 0, targetBitmap.BackBuffer + offset, length);
        }

        #endregion
        #region 영역 지우기 - ClearArea(rectangle)

        /// <summary>
        /// 영역 지우기
        /// </summary>
        /// <param name="rectangle">정수 사각형</param>
        private void ClearArea(Int32Rect rectangle)
        {
            int    bufferLength = 4 * rectangle.Width;
            byte[] lineBuffer   = new byte[bufferLength];

            for(int y = 0; y < rectangle.Height; y++)
            {
                int offset = (rectangle.Y + y) * this.stride + 4 * rectangle.X;

                CopyToBitmap(lineBuffer, this.bitmap, offset, bufferLength);
            }

            this.bitmap.AddDirtyRect(new Int32Rect(rectangle.X, rectangle.Y, rectangle.Width, rectangle.Height));
        }

        #endregion
        #region 영역 지우기 - ClearArea(rectangle)

        /// <summary>
        /// 영역 지우기
        /// </summary>
        /// <param name="rectangle">GIF 사각형 인터페이스</param>
        private void ClearArea(IGIFRectangle rectangle)
        {
            ClearArea(new Int32Rect(rectangle.Left, rectangle.Top, rectangle.Width, rectangle.Height));
        }

        #endregion
        #region 비트맵에서 복사하기 - CopyFromBitmap(targetByteArray, sourceBitmap, offset, length)

        /// <summary>
        /// 비트맵에서 복사하기
        /// </summary>
        /// <param name="targetByteArray">타겟 바이트 배열</param>
        /// <param name="sourceBitmap">소스 비트맵</param>
        /// <param name="offset">오프셋</param>
        /// <param name="length">길이</param>
        private static void CopyFromBitmap(byte[] targetByteArray, WriteableBitmap sourceBitmap, int offset, int length)
        {
            Marshal.Copy(sourceBitmap.BackBuffer + offset, targetByteArray, 0, length);
        }

        #endregion
        #region 이전 프레임 처분하기 - DisposePreviousFrame(currentFrame)

        /// <summary>
        /// 이전 프레임 처분하기
        /// </summary>
        /// <param name="currentFrame">현재 GIF 프레임</param>
        private void DisposePreviousFrame(GIFFrame currentFrame)
        {
            GIFGraphicControlExtension previousExtension = this.previousFrame?.GraphicControl;

            if(previousExtension != null)
            {
                switch(previousExtension.DisposalMethod)
                {
                    case GIFFrameDisposalMethod.None         :
                    case GIFFrameDisposalMethod.DoNotDispose :

                        break;

                    case GIFFrameDisposalMethod.RestoreBackground :

                        ClearArea(GetFixedUpFrameRectangle(this.previousFrame.Descriptor));

                        break;

                    case GIFFrameDisposalMethod.RestorePrevious :

                        CopyToBitmap(this.previousBackBuffer, this.bitmap, 0, this.previousBackBuffer.Length);

                        GIFLogicalScreenDescriptor descriptor = this.dataStream.Header.LogicalScreenDescriptor;

                        Int32Rect rectangle = new Int32Rect(0, 0, descriptor.Width, descriptor.Height);

                        this.bitmap.AddDirtyRect(rectangle);

                        break;
                }
            }

            GIFGraphicControlExtension extension = currentFrame.GraphicControl;

            if(extension != null && extension.DisposalMethod == GIFFrameDisposalMethod.RestorePrevious)
            {
                CopyFromBitmap(this.previousBackBuffer, this.bitmap, 0, this.previousBackBuffer.Length);
            }
        }

        #endregion
        #region 인터레이스 행 열거 가능형 구하기 - GetInterlacedRowEnumerable(height)

        /// <summary>
        /// 인터레이스 행 열거 가능형 구하기
        /// </summary>
        /// <param name="height"></param>
        /// <returns>인터레이스 행 열거 가능형</returns>
        private static IEnumerable<int> GetInterlacedRowEnumerable(int height)
        {
            /*
                Pass 1 : 행 0, 8, 16, 24...
                Pass 2 : 행 4, 12, 20, 28...
                Pass 3 : 행 2, 6, 10, 14...
                Pass 4 : 행 1, 3, 5, 7...
            */

            var passArray = new[]
            {
                new { Start = 0, Step = 8 },
                new { Start = 4, Step = 8 },
                new { Start = 2, Step = 4 },
                new { Start = 1, Step = 2 }
            };

            foreach(var pass in passArray)
            {
                int y = pass.Start;

                while(y < height)
                {
                    yield return y;

                    y += pass.Step;
                }
            }
        }

        #endregion
        #region 일반 행 열거 가능형 구하기 - GetNormalRowEnumerable(height)

        /// <summary>
        /// 일반 행 열거 가능형 구하기
        /// </summary>
        /// <param name="height">높이</param>
        /// <returns>일반 행 열거 가능형</returns>
        private static IEnumerable<int> GetNormalRowEnumerable(int height)
        {
            return Enumerable.Range(0, height);
        }

        #endregion
        #region 색상 쓰기 - WriteColor(lineBuffer, color, startIndex)

        /// <summary>
        /// 색상 쓰기
        /// </summary>
        /// <param name="lineBuffer">라인 버퍼</param>
        /// <param name="color">색상</param>
        /// <param name="startIndex">시작 인덱스</param>
        private static void WriteColor(byte[] lineBuffer, Color color, int startIndex)
        {
            lineBuffer[startIndex    ] = color.B;
            lineBuffer[startIndex + 1] = color.G;
            lineBuffer[startIndex + 2] = color.R;
            lineBuffer[startIndex + 3] = color.A;
        }

        #endregion
        #region 프레임 렌더링하기 (비동기) - RenderFrameAsync(frameIndex, cancellationToken)

        /// <summary>
        /// 프레임 렌더링하기 (비동기)
        /// </summary>
        /// <param name="frameIndex">프레임 인덱스</param>
        /// <param name="cancellationToken">취소 토큰</param>
        /// <returns>태스크</returns>
        private async Task RenderFrameAsync(int frameIndex, CancellationToken cancellationToken)
        {
            if(frameIndex < 0)
            {
                return;
            }

            GIFFrame frame = this.dataStream.FrameList[frameIndex];

            GIFImageDescriptor descriptor = frame.Descriptor;

            Int32Rect rectangle = GetFixedUpFrameRectangle(descriptor);

            using var indexStream = await GetIndexStreamAsync(frame, cancellationToken);

            using(this.bitmap.LockInScope())
            {
                if(frameIndex < this.previousFrameIndex)
                {
                    ClearArea(this.dataStream.Header.LogicalScreenDescriptor);
                }
                else
                {
                    DisposePreviousFrame(frame);
                }

                int    bufferLength = 4 * rectangle.Width;
                byte[] indexBuffer  = new byte[descriptor.Width];
                byte[] lineBuffer   = new byte[bufferLength];

                GIFPalette palette = this.paletteDictionary[frameIndex];

                int transparencyIndex = palette.TransparencyIndex ?? -1;

                IEnumerable<int> rowEnumerable = descriptor.Interlace ? GetInterlacedRowEnumerable(rectangle.Height) :
                                                                        GetNormalRowEnumerable(rectangle.Height);

                foreach(int y in rowEnumerable)
                {
                    indexStream.ReadAll(indexBuffer, 0, descriptor.Width);

                    int offset = (descriptor.Top + y) * this.stride + descriptor.Left * 4;

                    if(transparencyIndex >= 0)
                    {
                        CopyFromBitmap(lineBuffer, this.bitmap, offset, bufferLength);
                    }

                    for(int x = 0; x < rectangle.Width; x++)
                    {
                        byte index = indexBuffer[x];

                        int i = 4 * x;

                        if(index != transparencyIndex)
                        {
                            WriteColor(lineBuffer, palette[index], i);
                        }
                    }

                    CopyToBitmap(lineBuffer, this.bitmap, offset, bufferLength);
                }

                this.bitmap.AddDirtyRect(rectangle);
            }

            this.previousFrame      = frame;
            this.previousFrameIndex = frameIndex;
        }

        #endregion
        #region 실행하기 (비동기) - RunAsync(cancellationToken)

        /// <summary>
        /// 실행하기 (비동기)
        /// </summary>
        /// <param name="cancellationToken">취소 토큰</param>
        /// <returns>태스크</returns>
        private async Task RunAsync(CancellationToken cancellationToken)
        {
            while(true)
            {
                cancellationToken.ThrowIfCancellationRequested();

                Task<bool> timingTask = this.timingManager.NextAsync(cancellationToken);

                Task renderingTask = RenderFrameAsync(CurrentFrameIndex, cancellationToken);

                await Task.WhenAll(timingTask, renderingTask);

                if(!timingTask.Result)
                {
                    break;
                }

                CurrentFrameIndex = (CurrentFrameIndex + 1) % FrameCount;
            }
        }

        #endregion
    }
}