﻿using System;
using System.Diagnostics;
using System.IO;
using System.Runtime.CompilerServices;

using Buffer = System.Buffer;

namespace TestLibrary.Decompression
{
    /// <summary>
    /// LZW 압축 해제 스트림
    /// </summary>
    public class LZWDecompressStream : Stream
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Structure
        ////////////////////////////////////////////////////////////////////////////////////////// Internal

        #region 시퀀스 - Sequence

        /// <summary>
        /// 시퀀스
        /// </summary>
        internal struct Sequence
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 바이트 배열 - ByteArray

            /// <summary>
            /// 바이트 배열
            /// </summary>
            public byte[] ByteArray { get; }

            #endregion
            #region 클리어 코드 - IsClearCode

            /// <summary>
            /// 클리어 코드
            /// </summary>
            public bool IsClearCode { get; }

            #endregion
            #region 중단 코드 여부 - IsStopCode

            /// <summary>
            /// 중단 코드 여부
            /// </summary>
            public bool IsStopCode { get; }

            #endregion
            #region 클리어 코드 - ClearCode

            /// <summary>
            /// 클리어 코드
            /// </summary>
            public static Sequence ClearCode { get; } = new Sequence(true, false);

            #endregion
            #region 중단 코드 - StopCode

            /// <summary>
            /// 중단 코드
            /// </summary>
            public static Sequence StopCode { get; } = new Sequence(false, true);

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - Sequence(sourceByteArray)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="sourceByteArray"></param>
            public Sequence(byte[] sourceByteArray) : this()
            {
                ByteArray = sourceByteArray;
            }

            #endregion

            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 생성자 - Sequence(isClearCode, isStopCode)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="isClearCode">클리어 코드 여부</param>
            /// <param name="isStopCode">중단 코드 여부</param>
            private Sequence(bool isClearCode, bool isStopCode) : this()
            {
                IsClearCode = isClearCode;
                IsStopCode  = isStopCode;
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Method
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 추가하기 - Append(value)

            /// <summary>
            /// 추가하기
            /// </summary>
            /// <param name="value">값</param>
            /// <returns>시퀀스</returns>
            public Sequence Append(byte value)
            {
                byte[] byteArray = new byte[ByteArray.Length + 1];

                ByteArray.CopyTo(byteArray, 0);

                byteArray[ByteArray.Length] = value;

                return new Sequence(byteArray);
            }

            #endregion
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Class
        ////////////////////////////////////////////////////////////////////////////////////////// Internal

        #region 색상 테이블 - CodeTable

        /// <summary>
        /// 색상 테이블
        /// </summary>
        internal class CodeTable
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Field
            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region Field

            /// <summary>
            /// 최소 코드 길이
            /// </summary>
            private readonly int minimumCodeLength;

            /// <summary>
            /// 시퀀스 배열
            /// </summary>
            private readonly Sequence[] sequenceArray;

            /// <summary>
            /// 카운트
            /// </summary>
            private int count;

            /// <summary>
            /// 코드 길이
            /// </summary>
            private int codeLength;

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 인덱서 - this[index]

            /// <summary>
            /// 인덱서
            /// </summary>
            /// <param name="index">인덱스</param>
            /// <returns>시퀀스</returns>
            public Sequence this[int index]
            {
                [MethodImpl(MethodImplOptions.AggressiveInlining)]
                get
                {
                    return this.sequenceArray[index];
                }
            }

            #endregion
            #region 카운트 - Count

            /// <summary>
            /// 카운트
            /// </summary>
            public int Count
            {
                [MethodImpl(MethodImplOptions.AggressiveInlining)]
                get
                {
                    return this.count;
                }
            }

            #endregion
            #region 코드 길이 - CodeLength

            /// <summary>
            /// 코드 길이
            /// </summary>
            public int CodeLength
            {
                [MethodImpl(MethodImplOptions.AggressiveInlining)]
                get
                {
                    return this.codeLength;
                }
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - CodeTable(minimumCodeLength)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="minimumCodeLength">최소 코드 길이</param>
            public CodeTable(int minimumCodeLength)
            {
                this.minimumCodeLength = minimumCodeLength;
                this.codeLength        = this.minimumCodeLength + 1;

                int initialEntryCount = 1 << minimumCodeLength;

                this.sequenceArray = new Sequence[1 << MaximumCodeLength];

                for(int i = 0; i < initialEntryCount; i++)
                {
                    this.sequenceArray[this.count++] = new Sequence(new[] {(byte) i});
                }

                Add(Sequence.ClearCode);
                Add(Sequence.StopCode );
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Method
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 리셋하기 - Reset()

            /// <summary>
            /// 리셋하기
            /// </summary>
            [MethodImpl(MethodImplOptions.AggressiveInlining)]
            public void Reset()
            {
                this.count = (1 << this.minimumCodeLength) + 2;

                this.codeLength = this.minimumCodeLength + 1;
            }

            #endregion
            #region 추가하기 - Add(sequence)

            /// <summary>
            /// 추가하기
            /// </summary>
            /// <param name="sequence">시퀀스</param>
            [MethodImpl(MethodImplOptions.AggressiveInlining)]
            public void Add(Sequence sequence)
            {
                if(this.count >= this.sequenceArray.Length)
                {
                    return;
                }

                this.sequenceArray[this.count++] = sequence;

                if((this.count & (this.count - 1)) == 0 && this.codeLength < MaximumCodeLength)
                {
                    this.codeLength++;
                }
            }

            #endregion
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 최대 코드 길이
        /// </summary>
        private const int MaximumCodeLength = 12;

        /// <summary>
        /// 비트 리더
        /// </summary>
        private readonly BitReader reader;

        /// <summary>
        /// 코드 테이블
        /// </summary>
        private readonly CodeTable codeTable;

        /// <summary>
        /// 이전 코드
        /// </summary>
        private int previousCode;

        /// <summary>
        /// 잔여 바이트 배열
        /// </summary>
        private byte[] remainingByteArray;

        /// <summary>
        /// 스트림 끝 여부
        /// </summary>
        private bool endOfStream;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 읽기 가능 여부 - CanRead

        /// <summary>
        /// 읽기 가능 여부
        /// </summary>
        public override bool CanRead => true;

        #endregion
        #region 탐색 가능 여부 - CanSeek

        /// <summary>
        /// 탐색 가능 여부
        /// </summary>
        public override bool CanSeek => false;

        #endregion
        #region 쓰기 가능 여부 - CanWrite

        /// <summary>
        /// 쓰기 가능 여부
        /// </summary>
        public override bool CanWrite => true;

        #endregion
        #region 길이 - Length

        /// <summary>
        /// 길이
        /// </summary>
        public override long Length
        {
            get
            {
                throw new NotSupportedException();
            }
        }

        #endregion
        #region 위치 - Position

        /// <summary>
        /// 위치
        /// </summary>
        public override long Position
        {
            get
            {
                throw new NotSupportedException();
            }
            set
            {
                throw new NotSupportedException();
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - LZWDecompressStream(compressedByteArray, minimumCodeLength)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="compressedByteArray">압축 바이트 배열</param>
        /// <param name="minimumCodeLength">최소 코드 길이</param>
        public LZWDecompressStream(byte[] compressedByteArray, int minimumCodeLength)
        {
            this.reader    = new BitReader(compressedByteArray);
            this.codeTable = new CodeTable(minimumCodeLength);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region 검증하기 - Validate(sourceByteArray, offset, count)

        /// <summary>
        /// 검증하기
        /// </summary>
        /// <param name="sourceByteArray">소스 바이트 배열</param>
        /// <param name="offset">오프셋</param>
        /// <param name="count">카운트</param>
        [Conditional("DISABLED")]
        private static void Validate(byte[] sourceByteArray, int offset, int count)
        {
            if(sourceByteArray == null)
            {
                throw new ArgumentNullException(nameof(sourceByteArray));
            }

            if(offset < 0)
            {
                throw new ArgumentOutOfRangeException(nameof(offset), "Offset can't be negative");
            }

            if(count < 0)
            {
                throw new ArgumentOutOfRangeException(nameof(count), "Count can't be negative");
            }

            if(offset + count > sourceByteArray.Length)
            {
                throw new ArgumentException("Buffer is to small to receive the requested data");
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 버퍼 비우기 - Flush()

        /// <summary>
        /// 버퍼 비우기
        /// </summary>
        public override void Flush()
        {
        }

        #endregion
        #region 탐색하기 - Seek(offset, origin)

        /// <summary>
        /// 탐색하기
        /// </summary>
        /// <param name="offset">오프셋</param>
        /// <param name="origin">탐색 기준</param>
        /// <returns>위치</returns>
        public override long Seek(long offset, SeekOrigin origin)
        {
            throw new NotSupportedException();
        }

        #endregion
        #region 길이 설정하기 - SetLength(value)

        /// <summary>
        /// 길이 설정하기
        /// </summary>
        /// <param name="value">값</param>
        public override void SetLength(long value)
        {
            throw new NotSupportedException();
        }

        #endregion
        #region 읽기 - Read(targetByteArray, offset, count)

        /// <summary>
        /// 읽기
        /// </summary>
        /// <param name="targetByteArray">타겟 바이트 배열</param>
        /// <param name="offset">오프셋</param>
        /// <param name="count">카운트</param>
        /// <returns>값</returns>
        public override int Read(byte[] targetByteArray, int offset, int count)
        {
            Validate(targetByteArray, offset, count);

            if(this.endOfStream)
            {
                return 0;
            }

            int readCount = 0;

            FlushRemainingByteArray(targetByteArray, offset, count, ref readCount);

            while(readCount < count)
            {
                int code = this.reader.ReadBits(this.codeTable.CodeLength);
                
                if(!ProcessCode(code, targetByteArray, offset, count, ref readCount))
                {
                    this.endOfStream = true;

                    break;
                }
            }

            return readCount;
        }

        #endregion
        #region 쓰기 - Write(sourceByteArray, offset, count)

        /// <summary>
        /// 쓰기
        /// </summary>
        /// <param name="sourceByteArray">소스 바이트 배열</param>
        /// <param name="offset">오프셋</param>
        /// <param name="count">카운트</param>
        public override void Write(byte[] sourceByteArray, int offset, int count)
        {
            throw new NotSupportedException();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Public

        #region 바이트 배열 복사하기 - CopyByteArray(sourceByteArray, targetByteArray, offset, count, readCount)

        /// <summary>
        /// 바이트 배열 복사하기
        /// </summary>
        /// <param name="sourceByteArray">소스 바이트 배열</param>
        /// <param name="targetByteArray">타겟 바이트 배열</param>
        /// <param name="offset">오프셋</param>
        /// <param name="count">카운트</param>
        /// <param name="readCount">읽기 카운트</param>
        /// <returns>바이트 배열</returns>
        private static byte[] CopyByteArray(byte[] sourceByteArray, byte[] targetByteArray, int offset, int count, ref int readCount)
        {
            int byteCountToRead = Math.Min(sourceByteArray.Length, count - readCount);

            Buffer.BlockCopy(sourceByteArray, 0, targetByteArray, offset + readCount, byteCountToRead);

            readCount += byteCountToRead;

            byte[] remainingByteArray = null;

            if(byteCountToRead < sourceByteArray.Length)
            {
                int remainingBytesCount = sourceByteArray.Length - byteCountToRead;

                remainingByteArray = new byte[remainingBytesCount];

                Buffer.BlockCopy(sourceByteArray, byteCountToRead, remainingByteArray, 0, remainingBytesCount);
            }

            return remainingByteArray;
        }

        #endregion
        #region 잔여 바이트 배열 비우기 - FlushRemainingByteArray(targetByteArray, offset, count, readCount)

        /// <summary>
        /// 잔여 바이트 배열 비우기
        /// </summary>
        /// <param name="targetByteArray">타겟 바이트 배열</param>
        /// <param name="offset">오프셋</param>
        /// <param name="count">카운트</param>
        /// <param name="readCount">읽기 카운트</param>
        private void FlushRemainingByteArray(byte[] targetByteArray, int offset, int count, ref int readCount)
        {
            if(this.remainingByteArray != null)
            {
                this.remainingByteArray = CopyByteArray(this.remainingByteArray, targetByteArray, offset, count, ref readCount);
            }
        }

        #endregion
        #region 코드 테이블 초기화하기 - InitializeCodeTable()

        /// <summary>
        /// 코드 테이블 초기화하기
        /// </summary>
        private void InitializeCodeTable()
        {
            this.codeTable.Reset();

            this.previousCode = -1;
        }

        #endregion
        #region 코드 처리하기 - ProcessCode(code, targetByteArray, offset, count, readCount)

        /// <summary>
        /// 코드 처리하기
        /// </summary>
        /// <param name="code">코드</param>
        /// <param name="targetByteArray">타겟 바이트 배열</param>
        /// <param name="offset">오프셋</param>
        /// <param name="count">카운트</param>
        /// <param name="readCount">읽기 카운트</param>
        /// <returns>처리 결과</returns>
        private bool ProcessCode(int code, byte[] targetByteArray, int offset, int count, ref int readCount)
        {
            if(code < this.codeTable.Count)
            {
                LZWDecompressStream.Sequence sequence = this.codeTable[code];

                if(sequence.IsStopCode)
                {
                    return false;
                }

                if(sequence.IsClearCode)
                {
                    InitializeCodeTable();

                    return true;
                }

                this.remainingByteArray = CopyByteArray(sequence.ByteArray, targetByteArray, offset, count, ref readCount);

                if(this.previousCode >= 0)
                {
                    LZWDecompressStream.Sequence previousSequence = this.codeTable[this.previousCode];

                    LZWDecompressStream.Sequence newSequence = previousSequence.Append(sequence.ByteArray[0]);

                    this.codeTable.Add(newSequence);
                }
            }
            else
            {
                LZWDecompressStream.Sequence previousSequence = this.codeTable[this.previousCode];

                LZWDecompressStream.Sequence newSequence = previousSequence.Append(previousSequence.ByteArray[0]);

                this.codeTable.Add(newSequence);

                this.remainingByteArray = CopyByteArray(newSequence.ByteArray, targetByteArray, offset, count, ref readCount);
            }

            this.previousCode = code;

            return true;
        }

        #endregion
    }
}