﻿using System;
using System.IO;
using System.Threading;
using System.Threading.Tasks;

namespace TestLibrary.Extension
{
    /// <summary>
    /// 스트림 확장
    /// </summary>
    public static class StreamExtension
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 모두 읽기 (비동기) - ReadAllAsync(sourceStream, targetByteArray, offset, count, cancellationToken)

        /// <summary>
        /// 모두 읽기 (비동기)
        /// </summary>
        /// <param name="sourceStream">소스 스트림</param>
        /// <param name="targetByteArray">타겟 바이트 배열</param>
        /// <param name="offset">오프셋</param>
        /// <param name="count">카운트</param>
        /// <param name="cancellationToken">취소 토큰</param>
        /// <returns>태스크</returns>
        public static async Task ReadAllAsync
        (
            this Stream       sourceStream,
            byte[]            targetByteArray,
            int               offset,
            int               count,
            CancellationToken cancellationToken = default
        )
        {
            int totalReadCount = 0;

            while(totalReadCount < count)
            {
#if LACKS_STREAM_MEMORY_OVERLOADS

                int readCount = await sourceStream.ReadAsync
                (
                    targetByteArray,
                    offset + totalReadCount,
                    count - totalReadCount,
                    cancellationToken
                );

#else

                int readCount = await sourceStream.ReadAsync
                (
                    targetByteArray.AsMemory(offset + totalReadCount, count - totalReadCount),
                    cancellationToken
                );

#endif

                if(readCount == 0)
                {
                    throw new EndOfStreamException();
                }

                totalReadCount += readCount;
            }
        }

        #endregion
        #region 모두 읽기 - ReadAll(sourceStream, targetByteArray, offset, count)

        /// <summary>
        /// 모두 읽기
        /// </summary>
        /// <param name="sourceStream">소스 스트림</param>
        /// <param name="targetByteArray">타겟 바이트 배열</param>
        /// <param name="offset">오프셋</param>
        /// <param name="count">카운트</param>
        public static void ReadAll(this Stream sourceStream, byte[] targetByteArray, int offset, int count)
        {
            int totalReadCount = 0;

            while(totalReadCount < count)
            {
                int readCount = sourceStream.Read(targetByteArray, offset + totalReadCount, count - totalReadCount);

                if(readCount == 0)
                {
                    throw new EndOfStreamException();
                }

                totalReadCount += readCount;
            }
        }

        #endregion
        #region 바이트 읽기 (비동기) - ReadByteAsync(sourceStream, cancellationToken)

        /// <summary>
        /// 바이트 읽기 (비동기)
        /// </summary>
        /// <param name="sourceStream">소스 스트림</param>
        /// <param name="cancellationToken">취소 토큰</param>
        /// <returns>정수 태스크</returns>
        public static async Task<int> ReadByteAsync(this Stream sourceStream, CancellationToken cancellationToken = default)
        {
            byte[] targetByteArray = new byte[1];

#if LACKS_STREAM_MEMORY_OVERLOADS

            int readCount = await sourceStream.ReadAsync(targetByteArray, 0, 1, cancellationToken);
#else

            int readCount = await sourceStream.ReadAsync(targetByteArray.AsMemory(0, 1), cancellationToken);

#endif

            if(readCount == 0)
            {
                return -1;
            }

            return targetByteArray[0];
        }

        #endregion
        #region 버퍼 스트림 구하기 - AsBuffered(stream)

        /// <summary>
        /// 버퍼 스트림 구하기
        /// </summary>
        /// <param name="stream">스트림</param>
        /// <returns>버퍼 스트림</returns>
        public static Stream AsBuffered(this Stream stream)
        {
            if(stream is BufferedStream bufferedStream)
            {
                return bufferedStream;
            }

            return new BufferedStream(stream);
        }

        #endregion
        #region 복사하기 (비동기) - CopyToAsync(sourceStream, targetStream, progress, bufferSize, cancellationToken)

        /// <summary>
        /// 복사하기 (비동기)
        /// </summary>
        /// <param name="sourceStream">소스 스트림</param>
        /// <param name="targetStream">타겟 스트림</param>
        /// <param name="progress">진행 인터페이스</param>
        /// <param name="bufferSize">버퍼 크기</param>
        /// <param name="cancellationToken">취소 토큰</param>
        /// <returns>태스크</returns>
        public static async Task CopyToAsync
        (
            this Stream       sourceStream,
            Stream            targetStream,
            IProgress<long>   progress,
            int               bufferSize = 81920,
            CancellationToken cancellationToken = default
        )
        {
            byte[] byteArray       = new byte[bufferSize];
            int    byteCountRead;
            long   byteCountCopied = 0;

#if LACKS_STREAM_MEMORY_OVERLOADS

            while((byteCountRead = await sourceStream.ReadAsync(byteArray, 0, byteArray.Length, cancellationToken).ConfigureAwait(false)) != 0)
#else
            while((byteCountRead = await sourceStream.ReadAsync(byteArray.AsMemory(), cancellationToken).ConfigureAwait(false)) != 0)
#endif
            {
#if LACKS_STREAM_MEMORY_OVERLOADS

                await targetStream.WriteAsync(byteArray, 0, byteCountRead, cancellationToken).ConfigureAwait(false);

#else

                await targetStream.WriteAsync(byteArray.AsMemory(0, byteCountRead), cancellationToken).ConfigureAwait(false);

#endif

                byteCountCopied += byteCountRead;

                progress?.Report(byteCountCopied);
            }
        }

        #endregion
    }
}