﻿using System;
using System.IO;
using System.IO.Packaging;
using System.Linq;
using System.Net.Http;
using System.Security.Cryptography;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Resources;

using TestLibrary.Extension;

namespace TestLibrary
{
    /// <summary>
    /// URI 로더
    /// </summary>
    public class URILoader
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region URI에서 스트림 구하기 (비동기) - GetStreamFromURIAsync(uri, progress)

        /// <summary>
        /// URI에서 스트림 구하기 (비동기)
        /// </summary>
        /// <param name="uri">URI</param>
        /// <param name="progress">진행 인터페이스</param>
        /// <returns>스트림 태스크</returns>
        public static Task<Stream> GetStreamFromURIAsync(Uri uri, IProgress<int> progress)
        {
            if(uri.IsAbsoluteUri && (uri.Scheme == "http" || uri.Scheme == "https"))
            {
                return GetNetworkStreamAsync(uri, progress);
            }

            return GetStreamFromURICoreAsync(uri);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region 16진수 문자열 구하기 - ToHexadecimalString(sourceByteArray)

        /// <summary>
        /// 16진수 문자열 구하기
        /// </summary>
        /// <param name="sourceByteArray">소스 바이트 배열</param>
        /// <returns>16진수 문자열</returns>
        private static string ToHexadecimalString(byte[] sourceByteArray)
        {
            return sourceByteArray.Aggregate
            (
                new StringBuilder(),
                (stringBuilder, value) => stringBuilder.Append(value.ToString("X2")),
                stringBuilder => stringBuilder.ToString()
            );
        }

        #endregion
        #region 캐시 파일명 구하기 - GetCacheFileName(uri)

        /// <summary>
        /// 캐시 파일명 구하기
        /// </summary>
        /// <param name="uri">URI</param>
        /// <returns>캐시 파일명</returns>
        private static string GetCacheFileName(Uri uri)
        {
            using SHA1 sha1 = SHA1.Create();

            byte[] uriByteArray = Encoding.UTF8.GetBytes(uri.AbsoluteUri);

            byte[] hashByteArray = sha1.ComputeHash(uriByteArray);

            return ToHexadecimalString(hashByteArray);
        }

        #endregion
        #region 임시 파일 스트림 열기 (비동기) - OpenTemporaryFileStreamAsync(fileName)

        /// <summary>
        /// 임시 파일 스트림 열기 (비동기)
        /// </summary>
        /// <param name="fileName">파일명</param>
        /// <returns>스트림 태스크</returns>
        private static Task<Stream> OpenTemporaryFileStreamAsync(string fileName)
        {
            string path = Path.Combine(Path.GetTempPath(), fileName);

            Stream stream = null;

            try
            {
                stream = File.OpenRead(path);
            }
            catch(FileNotFoundException)
            {
            }

            return Task.FromResult(stream);
        }

        #endregion
        #region 임시 파일 스트림 생성하기 (비동기) - CreateTemporaryFileStreamAsync(fileName)

        /// <summary>
        /// 임시 파일 스트림 생성하기 (비동기)
        /// </summary>
        /// <param name="fileName">파일명</param>
        /// <returns>스트림 태스크</returns>
        private static Task<Stream> CreateTemporaryFileStreamAsync(string fileName)
        {
            string filePath = Path.Combine(Path.GetTempPath(), fileName);

            Stream stream = File.OpenWrite(filePath);

            stream.SetLength(0);

            return Task.FromResult(stream);
        }

        #endregion
        #region 임시 파일 삭제하기 (비동기) - DeleteTemporaryFileAsync(filePath)

        /// <summary>
        /// 임시 파일 삭제하기 (비동기)
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        /// <returns>태스크</returns>
        private static Task DeleteTemporaryFileAsync(string filePath)
        {
            if(File.Exists(filePath))
            {
                File.Delete(filePath);
            }

            return Task.FromResult(filePath);
        }

        #endregion
        #region 캐시 파일에 다운로드하기 (비동기) - DownloadToCacheFileAsync(uri, fileName, progress)

        /// <summary>
        /// 캐시 파일에 다운로드하기 (비동기)
        /// </summary>
        /// <param name="uri">URI</param>
        /// <param name="fileName">파일명</param>
        /// <param name="progress">진행 인터페이스</param>
        /// <returns>태스크</returns>
        private static async Task DownloadToCacheFileAsync(Uri uri, string fileName, IProgress<int> progress)
        {
            try
            {
                using HttpClient client = new HttpClient();

                HttpRequestMessage requestMessage = new HttpRequestMessage(HttpMethod.Get, uri);

                HttpResponseMessage responseMessage = await client.SendAsync(requestMessage);

                responseMessage.EnsureSuccessStatusCode();

                long length = responseMessage.Content.Headers.ContentLength ?? 0;

                using var responseStream = await responseMessage.Content.ReadAsStreamAsync();

                using var fileStream = await CreateTemporaryFileStreamAsync(fileName);

                IProgress<long> absoluteProgress = null;

                if(progress != null)
                {
                    absoluteProgress = new Progress<long>
                    (
                        byteCountCopied =>
                        {
                            if(length > 0)
                            {
                                progress.Report((int)(100 * byteCountCopied / length));
                            }
                            else
                            {
                                progress.Report(-1);
                            }
                        }
                    );
                }

                await responseStream.CopyToAsync(fileStream, absoluteProgress);
            }
            catch
            {
                await DeleteTemporaryFileAsync(fileName);

                throw;
            }
        }

        #endregion
        #region 네트워크 스트림 구하기 (비동기) - GetNetworkStreamAsync(uri, progress)

        /// <summary>
        /// 네트워크 스트림 구하기 (비동기)
        /// </summary>
        /// <param name="uri">URI</param>
        /// <param name="progress">진행 인터페이스</param>
        /// <returns>스트림 태스크</returns>
        private static async Task<Stream> GetNetworkStreamAsync(Uri uri, IProgress<int> progress)
        {
            string cacheFileName = GetCacheFileName(uri);

            Stream cacheStream = await OpenTemporaryFileStreamAsync(cacheFileName);

            if(cacheStream == null)
            {
                await DownloadToCacheFileAsync(uri, cacheFileName, progress);
            }

            progress.Report(100);

            return await OpenTemporaryFileStreamAsync(cacheFileName);
        }

        #endregion
        #region URI에서 스트림 구하기 (코어, 비동기) - GetStreamFromURICoreAsync(uri)

        /// <summary>
        /// URI에서 스트림 구하기 (코어, 비동기)
        /// </summary>
        /// <param name="uri">URI</param>
        /// <returns>스트림 태스크</returns>
        private static Task<Stream> GetStreamFromURICoreAsync(Uri uri)
        {
            if(uri.Scheme == PackUriHelper.UriSchemePack)
            {
                StreamResourceInfo info = uri.Authority == "siteoforigin:,,," ? Application.GetRemoteStream(uri) :
                                                                                Application.GetResourceStream(uri);

                if(info != null)
                {
                    return Task.FromResult(info.Stream);
                }

                throw new FileNotFoundException("Cannot find file with the specified URI");
            }

            if(uri.Scheme == Uri.UriSchemeFile)
            {
                return Task.FromResult<Stream>(File.OpenRead(uri.LocalPath));
            }

            throw new NotSupportedException("Only pack:, file:, http: and https: URIs are supported");
        }

        #endregion
    }
}